/***********************************************************************/
/*                                                                     */
/*   svm_learn.c                                                      */
/*                                                                     */
/*   Learning module of Support Vector Machine.                        */
/*                                                                     */
/*   Copyright: Thorsten Joachims                                      */
/*   Date: 08.12.97                                                    */
/*                                                                     */
/***********************************************************************/

# include <stdio.h>
# include <ctype.h>
# include <signal.h>
# include <math.h>
# include <string.h>
# include <stdlib.h> 
# include "svm.h"

char docfile[200];           /* file with training examples */
char modelfile[200];         /* file for resulting classifier */

double epsilon_crit;         /* tolerable error for distances used */
                             /* in stopping criterion */
double epsilon;              /* tolerable error for distances used for */ 
                             /* selecting examples for qp subproblem*/
double epsilon_a;            /* tolerable error on alphas */
double epsilon_const;        /* tolerable error on eq-constraint */
long opt_parts = 3;          /* keep n-1 parts of old qp problem and take */
                             /* one part of new examples */
long verbosity;              /* verbosity level (0-3) */

long   *kernel_cache_index;  /* cache some kernel evalutations */
double *kernel_cache_buffer; /* to improve speed */
long   *kernel_cache_occu;
long   kernel_cache_elems=0;
long   max_kernel_cache_elems=0;
long   kernel_cache_malloc();
int    kernel_cache_free(); 


double *optimize_qp();       /* interface to donlp2 */
double *opt_ce,*opt_ce0,*opt_g,*opt_g0,*opt_xinit; 
long opt_n;
double svm_c;

static int cmp_key_random(const long *, const long *);
int *randcmp;                /* sort key for cmp_key_random */
static int cmp_key_diff(const long *, const long *);  
double *diff;                /* sort key for cmp_key_diff */

long calculate_svm_model();
double estimate_error_bound();
double estimate_sphere();

int nx;


main (argc, argv)
int argc;
char *argv[];
{
  WORD **docs;  /* training examples */
  long *label,max_docs,max_words_doc,svm_maxqpsize,totwords,totdoc,ll;
  long remove_inconsistent;
  KERNEL_PARM kernel_parm;
  MODEL model;

  read_input_parameters(argc,argv,docfile,modelfile,&verbosity,
			&remove_inconsistent,&svm_maxqpsize,
			&svm_c,&epsilon_crit,&kernel_parm);
  epsilon      = epsilon_crit/10;
  epsilon_a    = 0.000000000002;
  epsilon_const= 0.0002;

  nol_ll(docfile,&max_docs,&max_words_doc,&ll); /* scan size of input file */
  max_words_doc+=2;
  ll+=2;

  docs = (WORD **)malloc(sizeof(WORD *)*max_docs);
  label = (long *)malloc(sizeof(long)*max_docs);

  read_documents(docfile,docs,label,max_words_doc,ll,&totwords,&totdoc);

  svm_learn(docs,label,totwords,totdoc,remove_inconsistent,svm_maxqpsize,
	    svm_c,&kernel_parm,&model);

  write_model(modelfile,&model,totwords,&kernel_parm);

  free(model.supvec);
  free(model.alpha);
  free(model.docnum);
  free(docs);
  free(label);
}


svm_learn(docs,label,totwords,totdoc,remove_inconsistent,svm_maxqpsize,svm_c,kernel_parm,model)
WORD **docs;
long *label,totwords,totdoc,remove_inconsistent,svm_maxqpsize;
double svm_c;
KERNEL_PARM *kernel_parm;
MODEL *model;
{
  long wpos,*inconsistent,*chosen,*key,i,j,k,new_inconsistents,wnum,pos;
  long inconsistentnum,choosenum,newchoosenum,choosenum1,choosenum2;
  long inner_region,outer_region,misclassified,supvecnum=0,upsupvecnum;
  double dist,len_w,criterion,maxdiff,*a_old,*lin,*dists,*a,c;

  inconsistent = (long *)malloc(sizeof(long)*(totdoc+2));
  chosen = (long *)malloc(sizeof(long)*(totdoc+2));
  key = (long *)malloc(sizeof(long)*(totdoc+2));
  a = (double *)malloc(sizeof(double)*(totdoc+2));
  a_old = (double *)malloc(sizeof(double)*(totdoc+2));
  lin = (double *)malloc(sizeof(double)*(totdoc+2));
  dists = (double *)malloc(sizeof(double)*(totdoc+2));
  diff = (double *)malloc(sizeof(double)*(totdoc+2));
  randcmp = (int *)malloc(sizeof(int)*(totdoc+2));
  model->supvec = (WORD **)malloc(sizeof(WORD *)*(totdoc+2));
  model->alpha = (double *)malloc(sizeof(double)*(totdoc+2));
  model->docnum = (long *)malloc(sizeof(long)*(totdoc+2));
  opt_ce = (double *)malloc(sizeof(double)*svm_maxqpsize);
  opt_ce0 = (double *)malloc(sizeof(double));
  opt_g = (double *)malloc(sizeof(double)*svm_maxqpsize*svm_maxqpsize);
  opt_g0 = (double *)malloc(sizeof(double)*svm_maxqpsize);
  opt_xinit = (double *)malloc(sizeof(double)*svm_maxqpsize);

  max_kernel_cache_elems=svm_maxqpsize;
  kernel_cache_index = (long *)malloc(sizeof(long)*totdoc);
  kernel_cache_occu = (long *)malloc(sizeof(long)*max_kernel_cache_elems);
  kernel_cache_buffer = (double *)malloc(sizeof(double)*max_kernel_cache_elems*totdoc);

  model->sv_num=0;

  kernel_cache_elems=0;   /* initialize cache */
  for(i=0;i<totdoc;i++) {
    kernel_cache_index[i]=-1;
  }
  for(i=0;i<max_kernel_cache_elems;i++) {
    kernel_cache_occu[i]=0;
  }

  for(i=0;i<totdoc;i++) {    /* various inits */
    inconsistent[i]=0;
    chosen[i]=0;
    a[i]=0;
    a_old[i]=0;
    key[i]=i;
    lin[i]=0;
  }
  inconsistentnum=0;
  newchoosenum=0;
  choosenum=0;
  choosenum1=0;
  choosenum2=0;
  misclassified=0;
  inner_region=0;
                             /* select examples for initial qp subproblem */
  for(c=0;c<totdoc;c+=opt_parts*totdoc/(double)min(svm_maxqpsize,totdoc)) {
    i=(int)c;
    chosen[i]=1;
    choosenum++;
    if(label[i] == 1) {
      choosenum1++;
    }
    else {
      choosenum2++;
    }
    newchoosenum++;
  }
  for(i=0;(choosenum1<5) && (i<totdoc);i++) { /* make sure that there */
    if((label[i] == 1) && (!chosen[i])) {     /* are enough +1 examples */
      chosen[i]=1;                            /* in first training set */
      choosenum++;
      choosenum1++;
      newchoosenum++;
    }
  }
  for(i=0;(choosenum2<5) && (i<totdoc);i++) {  /* make sure that there */
    if((label[i] == -1) && (!chosen[i])) {     /* are enough -1 examples */
      chosen[i]=1;                             /* in first training set */
      choosenum++;
      choosenum2++;
      newchoosenum++;
    }
  }				
  misclassified=0;
  inner_region=1;
  outer_region=1;
                            /* repeat this loop until we have convergence */
  for(j=1;((inner_region>0) || (outer_region>0));j++) {
    if(verbosity>=2) {
      printf(
	"Optimization pass number %d: %d (%d pos/%d neg) examples chosen:\n"
	,j,choosenum,choosenum1,choosenum2); fflush(stdout);
    }
    else if(verbosity==1) {
      printf("."); fflush(stdout);
    }

    optimize_svm(docs,label,chosen,model,totdoc,a,svm_c,kernel_parm);
    supvecnum=calculate_svm_model(docs,label,a,svm_c,chosen,totdoc,totwords,
				  model,kernel_parm);

    if(verbosity>=3) {
      /* Calculate some non essential values for verbose output */
      /* Watch out, this may be pretty costly to compute!!! */
      len_w=model_length_s(model,kernel_parm);
      printf("Current model length |w|=%.5f\n",len_w);
      criterion=0;
      for(i=0;i<totdoc;i++) {
	criterion+=a[i];
      }
      criterion=-criterion+0.5*len_w*len_w;
      printf("Criterion is at W(a)=%.20f \n",criterion);
    }

    if(verbosity>=3) {
      printf("Classifying training examples..."); fflush(stdout);
    }
    update_linear_component(docs,label,a,a_old,chosen,totdoc,kernel_parm,lin);

    for(i=0;i<totdoc;i++) {
      a_old[i]=a[i];
      if(chosen[i]) { 
	chosen[i]++;
      }
    }
    choosenum=0;
    choosenum1=0;
    choosenum2=0;
    newchoosenum=0;
    misclassified=0;
    inner_region=0;
    outer_region=0;
    maxdiff=0;
    new_inconsistents=0;
    for(i=0;i<totdoc;i++) {
      randcmp[i]=rand(); /* for random ordering */
      dists[i]=(lin[i]-model->b)*(double)label[i];/* distance from hyperplane*/
      diff[i]=fabs(dists[i]-1.0); /* violation of KT-constraints */
      /* the ordering induced by diff[i] is used to select the examples */
      /* for the next qp subproblem. the examples which 'violate the */
      /* KT-constraints most' are selected first. */
    }
    for(k=0;k<totdoc;k++) {
      key[k]=k;
    }
    if((j % 20) < 19) {  
      qsort((long *)key,              /* the examples which 'violate the KT- */
	    totdoc,   /* constraints most' are selected first. */ 
	    sizeof(long),
	    (int (*)(const void*, const void*))cmp_key_diff); 
    }
    else {
      qsort(key,              /* random ordering, just to avoid getting */ 
	    (size_t)totdoc,   /* stuck for whatever silly numerical reasons */
	    sizeof(long),
	    (int (*)(const void*, const void*))cmp_key_random); 
    }
    for(k=0;k<totdoc;k++) {
      i=key[k];
      if(!inconsistent[i]) {
	dist=dists[i];
	if((diff[i]>maxdiff)  /* largest violation */
	   && (((dist<1) && (a[i]<(svm_c-epsilon_a)))
	       || ((dist>1) && (a[i]>(0+epsilon_a))))) {
	  maxdiff=diff[i];
	}
	if(dist <= 0) {       /* counts for termination criterion */
	  misclassified++;
	}
	else if((a[i]<(svm_c-epsilon_a)) && (dist < (1-epsilon_crit))) {
	  inner_region++;
	}
	else if((a[i]>0+epsilon_a) && (dist > (1+epsilon_crit))) {
	  outer_region++;
	}
	/* pick examples for next qp subproblem */
	if((a[i]>=(svm_c-epsilon_a)) /* multiplier a[i] at upper bound, but */ 
	   && (dist<=(1+epsilon))){  /* still distance smaller 1 */
	  if(remove_inconsistent) {  /* remove inconsitent examples? */
	    inconsistent[i]=1;   /* never choose again */
	    inconsistentnum++;
	    new_inconsistents=1; /* throw out and start over */
	    if(verbosity>=3) {
	      printf("inconsistent(%ld,%ld)..",i,label[i]); fflush(stdout);
	    }
	  }
	  chosen[i]=0;
	}
	else if((a[i]>=(svm_c-epsilon_a)) /* a[i] at upper bound, but this */
		&& (dist>(1+epsilon))){   /* is too much */
	  chosen[i]=1;
	  choosenum++;
	  if(label[i]==1) {
	    choosenum1++;
	  }
	  else {
	    choosenum2++;
	  }
	}
	else if((!chosen[i])            /* potential support vector, but */
		&& (a[i]>(0+epsilon_a)) /* distance not equal to 1 */
		&& ((dist<(1-epsilon)) || (dist>(1+epsilon))))  {   
	  chosen[i]=1;
	  choosenum++;
	  if(label[i]==1) {
	    choosenum1++;
	  }
	  else {
	    choosenum2++;
	  }
	}		
	else if((a[i]<=(0+epsilon_a))     /* closer than current support */
		&& (dist<(1-epsilon))) {  /* vectors */
	  chosen[i]=1;
	  choosenum++;
	  if(label[i]==1) {
	    choosenum1++;
	  }
	  else {
	    choosenum2++;
	  }
	}	       
	else if(a[i]<=(0+epsilon_a)) {    /* nothing to do here */
	  chosen[i]=0;
	}
      }
    }    

    for(k=totdoc-1;(choosenum1>(svm_maxqpsize/(2*opt_parts))) && (k>=0);k--) {
      i=key[k];
      if((chosen[i] == 1) && (label[i] == 1)) { /* remove some new +1 */
	chosen[i]=0;                            /* examples from training */
	choosenum--;                            /* set to get equal number */
	choosenum1--;                           /* as -1 examples */
	newchoosenum--;
      }
    }
    for(k=totdoc-1;(choosenum2>(svm_maxqpsize/(2*opt_parts))) && (k>=0);k--) {
      i=key[k];
      if((chosen[i] == 1) && (label[i] == -1)) { /* remove some new -1 */
	chosen[i]=0;                             /* examples from training */
	choosenum--;                             /* set so that qp filled up */
	choosenum2--;
	newchoosenum--;
      }
    }

    for(i=0;i<totdoc;i++) {                      /* add n-th part of old ex */
      if((chosen[i] >= 2) && (chosen[i] <= opt_parts)) {
	choosenum++; 
	newchoosenum++;
	if(label[i]==1) {
	  choosenum1++;
	}
	else {
	  choosenum2++;
	}
      }
      else if(chosen[i] > opt_parts) {
	chosen[i]=0;
      }
    }

    if(remove_inconsistent && new_inconsistents) {  /* start tabula rasa */
      for(i=0;i<totdoc;i++) {
	a[i]=0;
	a_old[i]=0;
	lin[i]=0;
	chosen[i]=0;
	choosenum1=0;
	choosenum2=0;
	choosenum=0;
      }
    }

    for(i=0;(choosenum1<5) && (i<totdoc);i++) { /* make sure that there */
      if((label[i] == 1) && (!chosen[i])        /* are enough +1 examples */
         && (inconsistent[i] == 0)) {           /* in the training set */
	chosen[i]=1;                            
	choosenum++;
	choosenum1++;
	newchoosenum++;
      }
    }
    for(i=0;(choosenum2<5) && (i<totdoc);i++) {  /* make sure that there */
      if((label[i] == -1) && (!chosen[i])        /* are enough -1 examples */
         && (inconsistent[i] == 0)) {            /* in the training set */
	chosen[i]=1;                             
	choosenum++;
	choosenum2++;
	newchoosenum++;
      }
    }

    if(verbosity>=3) {
      printf("done\n");
    }
    if(verbosity>=2) {
      printf("Results of pass %d: (%d misclassified, %d in inner region, %d in outer region, %d inconsistent examples, maxdiff=%.5f, supvecnum=%d)\n",j,misclassified,inner_region,outer_region,inconsistentnum,maxdiff,supvecnum); fflush(stdout);
    }
  }
  if(verbosity>=1) {
    if(verbosity==1) printf("done. (%d iterations)\n",j);
    printf("Optimization finished: (%d misclassified, %d in inner region, %d in outer region, %d inconsistent examples, maxdiff=%.5f, supvecnum=%d):\n",misclassified,inner_region,outer_region,inconsistentnum,maxdiff,supvecnum); 
    if(remove_inconsistent) {

      printf("Number of SV: %ld (plus %ld inconsistent examples)\n",supvecnum,inconsistentnum);
    }
    else {
      upsupvecnum=0;
      for(i=1;i<model->sv_num;i++) {
	if(fabs(model->alpha[i]) > (svm_c-epsilon_a)) upsupvecnum++;
      }
      printf("Number of SV: %ld (including %ld at upper bound)\n",supvecnum,upsupvecnum);
    }

    /* Next line uses bound (5.23) from [Vapnik/95]. The number of SV */
    /* we get in our experiment is assumed to be the expected number of */
    /* SV for the task. */
    /* printf("Expected accurary estimated to be above %.3f (given separability)\n", (double)1 - ((double)supvecnum / (totdoc+1))); */
    /* printf("Probability of error below %.3f with 95-confidence (given separability)\n", estimate_error_bound(model,totwords,docs,totdoc,(double)0.95,kernel_parm)); */
    estimate_error_bound(model,totwords,docs,totdoc,(double)0.95,kernel_parm);
  }

  free(inconsistent);
  free(chosen);
  free(key);
  free(a);
  free(a_old);
  free(lin);
  free(dists);
  free(diff);
  free(randcmp);
  free(opt_ce);
  free(opt_ce0);
  free(opt_g);
  free(opt_g0);
  free(opt_xinit);
  free(kernel_cache_index);
  free(kernel_cache_occu);
  free(kernel_cache_buffer);
}
 
static int cmp_key_diff(a,b)  /* sort in decreasing order */
const long *a,*b;
{
  if(diff[*a]>diff[*b]) {
    return(-1);
  }
  else if(diff[*a]<diff[*b]) {
    return(1);
  }
  else {
    return(0);
  }
}

int cmp_key_random(a,b)    /* sort in random order (uses randcmp[]) */
const long *a,*b;
{
  if(randcmp[*a]<randcmp[*b]) {
    return(-1);
  }
  else if(randcmp[*a]>randcmp[*b]) {
    return(1);
  }
  else {
    return(0);
  }
}

update_linear_component(docs,label,a,a_old,chosen,totdoc,kernel_parm,lin)
WORD **docs;
long *label;
double *a,*a_old;                 
long *chosen,totdoc;
KERNEL_PARM *kernel_parm;
double *lin;
{
  register long i,j;
  register double *aicache,adiff;

  for(i=0;i<totdoc;i++) {     /* clean up cache a little */
    if((!chosen[i]) && (kernel_cache_index[i] != -1)) {
      kernel_cache_free(kernel_cache_index[i]); 
      kernel_cache_index[i]=-1;
    }
  }
  for(i=0;i<totdoc;i++) {
    if((chosen[i]) && (a[i] != a_old[i])) {
      adiff=(a[i]-a_old[i])*(double)label[i];
      if(kernel_cache_index[i] == -1) {      
	if((kernel_cache_index[i] = kernel_cache_malloc()) >= 0) {
	  aicache=(double *)((long)kernel_cache_buffer
	          +(totdoc*sizeof(double)*kernel_cache_index[i]));
	  for(j=0;j<totdoc;j++) {  /* fill cache */
	    aicache[j]=kernel(kernel_parm,docs[i],docs[j]);
	    lin[j]+=adiff*aicache[j];
	  }
	}
	else { /* cache full */
	  for(j=0;j<totdoc;j++) {
	    lin[j]+=adiff*kernel(kernel_parm,docs[i],docs[j]);
	  }
	}
      }
      else {   /* use cached values */
	aicache=(double *)((long)kernel_cache_buffer
	        +(totdoc*sizeof(double)*kernel_cache_index[i]));
	for(j=0;j<totdoc;j++) {
	  lin[j]+=adiff*aicache[j];
	}
      }
    }
  }
}

long kernel_cache_malloc() 
{
  long i;

  if(kernel_cache_elems < max_kernel_cache_elems) {
    for(i=0;i<max_kernel_cache_elems;i++) {
      if(!kernel_cache_occu[i]) {
	kernel_cache_occu[i]=1;
	kernel_cache_elems++;
	return(i);
      }
    }
  }
  return(-1);
}

kernel_cache_free(i)
long i;
{
  kernel_cache_occu[i]=0;
  kernel_cache_elems--;
}
  
write_model(modelfile,model,totwords,kernel_parm)
char *modelfile;
MODEL *model;
long totwords;
KERNEL_PARM *kernel_parm;
{
  FILE *modelfl;
  long j,i;

  if(verbosity>=1) {
    printf("Writing model file..."); fflush(stdout);
  }
  if ((modelfl = fopen (modelfile, "w")) == NULL)
  { perror (modelfile); exit (1); }
  fprintf(modelfl,"%ld\n",kernel_parm->kernel_type);
  fprintf(modelfl,"%ld\n",kernel_parm->poly_degree);
  fprintf(modelfl,"%.8g\n",kernel_parm->rbf_gamma);
  fprintf(modelfl,"%.8g\n",kernel_parm->sigmoid_gamma);
  fprintf(modelfl,"%.8g\n",kernel_parm->sigmoid_c);
 
  fprintf(modelfl,"%ld\n",model->sv_num);
  fprintf(modelfl,"%.8g\n",model->b);
  for(i=1;i<model->sv_num;i++) {
    fprintf(modelfl,"%.8g ",model->alpha[i]);
    for (j=0; (model->supvec[i])[j].wnum; j++) {
      fprintf(modelfl,"%ld:%.8g ",(model->supvec[i])[j].wnum,(model->supvec[i])[j].weight);
    }
    fprintf(modelfl,"\n");
  }
  fclose(modelfl);
  if(verbosity>=1) {
    printf("done\n");
  }
}

double estimate_error_bound(model,dimensions,docs,examplenum,delta,kernel_parm) 
MODEL *model;
long dimensions,examplenum;
double delta;
KERNEL_PARM *kernel_parm;
{
  double h,w,R;

  /* follows chapter 5.6.4 in [Vapnik/95] */

  w=model_length_s(model,kernel_parm);
  R=estimate_sphere(docs,examplenum,dimensions,kernel_parm); 
  if(verbosity>=1) {
    fprintf(stdout,"Norm of weight vector: |w|=%.5f\n",w);
  }
  h = w*w * R*R; 
  if(verbosity>=1) {
    fprintf(stdout,"Estimated VCdim of classifier: VCdim<=%.5f\n",h);
  }
  return(4*((h * (log(2*examplenum/h) + 1)) - log((1-delta)/4))/examplenum);
}

double estimate_sphere(docs,totdoc,totwords,kernel_parm) 
WORD **docs;          /* approximates the radius of the ball containing the */
long totdoc,totwords; /* training data by bounding it with the length of */
KERNEL_PARM *kernel_parm; /* the longest training example */
{
  register long j;
  double xlen,maxxlen=0;
  WORD nulldoc;

  if(verbosity>=1) {
    fprintf(stdout,"Estimating radius R of sphere..."); fflush(stdout);
  }
  nulldoc.wnum=0;
  for(j=0;j<totdoc;j++) {
    xlen=sqrt(kernel(kernel_parm,docs[j],docs[j])
	      -2*kernel(kernel_parm,docs[j],&nulldoc)
	      +kernel(kernel_parm,&nulldoc,&nulldoc));
    if(xlen>maxxlen) {
      maxxlen=xlen;
    }
  }
  if(verbosity>=1) {
    fprintf(stdout,"done (R <= longest training vector = %.5f)\n",maxxlen);
  }
  
  return(maxxlen);
}

compute_matrices_for_optimization(docs,label,chosen,key,model,a,varnum,totdoc,svm_c,kernel_parm)
WORD **docs;
long *label,*chosen,*key;
MODEL *model;
double *a;
long varnum,totdoc;
double svm_c;
KERNEL_PARM *kernel_parm;
{
  /* writes the global variables opt_ce, opt_ce0, opt_g, opt_g0 */

  register long ki,i,j,k;
  register double qbn,init,yn;
  WORD *ex;

  if(verbosity>=3) {
    fprintf(stdout,"Computing optimization matrices (type %ld kernel [degree %ld, rbf_gamma %lf, sigmoid_gamma %lf, sigmoid_c %lf])...",kernel_parm->kernel_type,kernel_parm->poly_degree,kernel_parm->rbf_gamma,kernel_parm->sigmoid_gamma,kernel_parm->sigmoid_c); 
    fflush(stdout);
  }
  
  opt_n=varnum;

  /* Compute the matrix for equality constraints */
  for(i=0;i<varnum;i++) {
    opt_ce[i]=label[key[i]];
  }
  opt_ce0[0]=0; /* compute the constant */
  for(i=0;i<totdoc;i++) { 
    if(!chosen[i]) {
      opt_ce0[0]+=a[i]*(double)label[i];
    }
  }
  yn=-opt_ce0[0];

  /* compute linear part of objective function */
  for(i=0;i<varnum;i++) {
    ex=docs[key[i]];
    qbn=0;
    for(j=1;j<model->sv_num;j++) { /* start at 1 */
      if(!chosen[model->docnum[j]]) {
	qbn+=kernel(kernel_parm,model->supvec[j],ex)*model->alpha[j];
      }
    }
    opt_g0[i]=-1+qbn*label[key[i]];
  }

  /* compute quadratic part of objective function */
  for(i=0;i<varnum;i++) {
    ki=key[i];
    for(j=i;j<varnum;j++) {
      opt_g[varnum*i+j]=label[ki]*label[key[j]]*kernel(kernel_parm,docs[ki],docs[key[j]]);
    }
    if(verbosity>=3) {
      if(i % 20 == 0) {
	fprintf(stdout,"%d..",i); fflush(stdout);
      }
    }
  }

  /* assure starting at feasable point */
  for(i=0;i<varnum;i++) {
    init=0;
    if((yn * (double)label[key[i]]) > 0) {         /* assure starting at */
      if((yn * (double)label[key[i]]) > svm_c) {   /* feasable point */
	init=svm_c;
	yn-=svm_c * (double)label[key[i]];
      }
      else {
	init=yn * (double)label[key[i]];
	yn=0;
      }
    }
    opt_xinit[i]=init;
  }

  if(verbosity>=3) {
    fprintf(stdout,"done\n");
  }
}

optimize_svm(docs,label,chosen,model,totdoc,a,svm_c,kernel_parm)
WORD **docs;
long *label,*chosen;
MODEL *model;
long totdoc;
double *a,svm_c;
KERNEL_PARM *kernel_parm;
{
    long i,j,*varnum2dnum,varnum,anum;
    double *a_v,constraint;

    varnum=0;
    varnum2dnum = (long *)malloc(sizeof(long)*(totdoc+1));
    for(i=0;i<totdoc;i++) {
      if(chosen[i]) {
	varnum2dnum[varnum]=i;
	varnum++;
      }
    }
    varnum2dnum[varnum]=-1;

    compute_matrices_for_optimization(docs,label,chosen,varnum2dnum,model,
				      a,varnum,totdoc,svm_c,kernel_parm);

    if(verbosity>=3) {
      printf("Running optimizer..."); fflush(stdout);
    }
    a_v=optimize_qp(a_v);      
    if(verbosity>=3) {         
      printf("done\n");
    }

    constraint=opt_ce0[0];           /* check, if the equality constraint */
    for(i=0;i<varnum;i++) {          /* holds. This tells us whether the */
      constraint+=opt_ce[i]*a_v[i];  /* optimizer converged properly. */
    }
    if(fabs(constraint) > epsilon_const) {
      printf("\nError: Optimization did not converge! EQ-Constraint violated: %lf\n",constraint);
    }

    for(i=0;i<varnum;i++) {
      if(a_v[i]>(0+epsilon_a)) {
	a[varnum2dnum[i]]=a_v[i];
      }
      else {
	a[varnum2dnum[i]]=0;
      }
    }

    free(varnum2dnum);
}

long calculate_svm_model(docs,labelv,a,svm_c,chosen,totdoc,numw,model,kernel_parm)
WORD **docs;
long *labelv;
double *a,svm_c;
long *chosen,totdoc,numw;
MODEL *model;
KERNEL_PARM *kernel_parm;
{
  long supvecnum,supvecnum_old=0;
  WORD *pdoc=0,*doc;
  long label,i,j,w;
  double pval=0;

  if(verbosity>=3) {
    printf("Calculating model..."); fflush(stdout);
  }
  supvecnum=0;
  model->supvec[0]=0;  /* element 0 reserved and empty for now */
  model->alpha[0]=0;
  model->docnum[0]=-1;
  supvecnum++;

  for(i=0;i<totdoc;i++) {
    if (a[i]>(0+epsilon_a)) {		/* support vector */
      model->supvec[supvecnum]=docs[i];
      model->alpha[supvecnum]=a[i]*labelv[i];
      model->docnum[supvecnum]=i;
      supvecnum++;
      if((!pdoc) && (chosen[i])                   /* save this for */
	 && (a[i]<(svm_c-epsilon_a))) {           /* calculating b later */
	pdoc=docs[i]; 
	pval=(double)labelv[i];
      }
    }
    if(verbosity>=3) {
      if((supvecnum % 100 == 0) && (supvecnum>supvecnum_old)) {
	supvecnum_old=supvecnum;
	printf("%ld..",supvecnum); fflush(stdout);
      }
    }
  }
  model->sv_num=supvecnum;
  if(pdoc) {
    model->b=0;	       /* now calculate threshold b */
    model->b=-pval+classify_example(model,pdoc,kernel_parm);
  }
  else {
    printf("Warning: No non-zero multiplier NOT at upper bound => set C higher!\n");
  }
  if(verbosity>=3) {
    printf("done (%ld support vectors)\n",supvecnum-1);
  }
  return(supvecnum-1); /* have to substract one, since element 0 is empty */
}


read_input_parameters(argc,argv,docfile,modelfile,verbosity,
		      remove_inconsistent,svm_maxqpsize,svm_c,
		      epsilon_crit,kernel_parm)
int argc;
char *argv[],*docfile,*modelfile;
long *verbosity,*remove_inconsistent,*svm_maxqpsize;
double *svm_c,*epsilon_crit;
KERNEL_PARM *kernel_parm;
{
  long i;
  
  /* set default */
  strcpy (modelfile, "svm_model");
  (*verbosity)=2;
  (*remove_inconsistent)=1;
  (*svm_maxqpsize)=100;
  (*svm_c)=1000;
  (*epsilon_crit)=0.001;
  kernel_parm->kernel_type=0;
  kernel_parm->poly_degree=3;
  kernel_parm->rbf_gamma=1.0;
  kernel_parm->sigmoid_gamma=2;
  kernel_parm->sigmoid_c=1;

  for(i=1;(i<argc) && ((argv[i])[0] == '-');i++) {
    switch ((argv[i])[1]) 
      { 
      case 'h': print_help(); exit(0);
      case 'v': i++; (*verbosity)=atol(argv[i]); break;
      case 'i': i++; (*remove_inconsistent)=atol(argv[i]); break;
      case 'q': i++; (*svm_maxqpsize)=atol(argv[i]); break;
      case 'c': i++; (*svm_c)=atof(argv[i]); break;
      case 'e': i++; (*epsilon_crit)=atof(argv[i]); break;
      case 't': i++; kernel_parm->kernel_type=atol(argv[i]); break;
      case 'd': i++; kernel_parm->poly_degree=atol(argv[i]); break;
      case 'g': i++; kernel_parm->rbf_gamma=atof(argv[i]); break;
      case 's': i++; kernel_parm->sigmoid_gamma=atof(argv[i]); break;
      case 'r': i++; kernel_parm->sigmoid_c=atof(argv[i]); break;
      default: printf("\nUnrecognized option %s!\n\n",argv[i]);
	       print_help();
	       exit(0);
      }
  }
  if(i>=argc) {
    printf("\nNot enough input parameters!\n\n");
    print_help();
    exit(0);
  }
  strcpy (docfile, argv[i]);
  if((i+1)<argc) {
    strcpy (modelfile, argv[i+1]);
  }
}

print_help()
{
  printf("\nSVM-light V0.9: Support Vector Machine, learning module     11.11.97\n");
  copyright_notice();
  printf("   usage: svm_learn [options] example_file model_file\n\n");
  printf("options: -h          -> this help\n");
  printf("         -v [0..3]   -> verbosity level (default 2)\n");
  printf("         -i [0,1]    -> remove inconsistent training examples\n");
  printf("                        and retrain (default 1)\n");
  printf("         -q [50..%d]-> maximum size of QP-subproblems (default 100)\n",nx);
  printf("         -c float    -> C: trade-off between training error\n");
  printf("                        and margin (default 1000)\n");
  printf("         -e float    -> eps: Allow that error when fitting constraints\n");
  printf("                        y [w*x+b] >=1-eps (default 0.001)\n");
  printf("         -t int      -> type of kernel function:\n");
  printf("                        0: linear\n");
  printf("                        1: polynomial (a*b+1)^d\n");
  printf("                        2: radial basis function exp(-gamma ||a-b||^2)\n");
  printf("                        3: sigmoid tanh(s a*b + c)\n");
  printf("         -d int      -> parameter d in polynomial kernel\n");
  printf("         -g float    -> parameter gamma in rbf kernel\n");
  printf("         -s float    -> parameter s in sigmoid kernel\n");
  printf("         -r float    -> parameter c in sigmoid kernel\n\n");
}


